// -*- C++ -*-
/*!
 * @file  RMRC.cpp
 * @brief 分解運動速度制御
 * @date $Date$
 *
 * $Id$
 */

#include "RMRC.h"

// Module specification
// <rtc-template block="module_spec">
static const char* rmrc_spec[] =
  {
    "implementation_id", "RMRC",
    "type_name",         "RMRC",
    "description",       "分解運動速度制御",
    "version",           "1.0.0",
    "vendor",            "FSI",
    "category",          "RMRC",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.n_size", "7",
    "conf.default.m_max", "100",
    // Widget
    "conf.__widget__.n_size", "text",
    "conf.__widget__.m_max", "text",
    // Constraints
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RMRC::RMRC(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_thIn("Th", m_th),
    m_velOut("Vel", m_vel),
    m_ComRMRCPort("ComRMRC")

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
RMRC::~RMRC()
{
}



RTC::ReturnCode_t RMRC::onInitialize()
{
  m_eth=new ExecTh(this);
  m_thIn.addConnectorDataListener(ON_BUFFER_WRITE, m_eth);
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("Th", m_thIn);
  
  // Set OutPort buffer
  addOutPort("Vel", m_velOut);
  
  // Set service provider to Ports
  m_ComRMRCPort.registerProvider("com_frm_ctrl", "ComFrmCtrl", m_com_frm_ctrl);
  m_ComRMRCPort.registerProvider("com_fk", "ComFk", m_com_fk);
  m_ComRMRCPort.registerProvider("com_move", "ComMove", m_com_move);
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  addPort(m_ComRMRCPort);
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("n_size", m_n_size, "7");
  bindParameter("m_max", m_m_max, "100");
  
  m_vel.data.length(m_n_size);
  // </rtc-template>

  /*  
   *  frm_ctrl code start
   *
   * default
   * k_gain = 1.0, means time constant is 1 sec.
   * v_max = 200[mm/sec]
   * k_pos_rot = 200[mm], means that sizes of objects are about this size.
   */
  m_com_frm_ctrl.set_param(1.0, 200, 200);
  m_com_frm_ctrl.settle = 1.0;
  m_com_frm_ctrl.state = 1;
  
  m_com_move.ref_flag = 0;  //目標手先位置・姿勢未入力
  m_cur_flag = 0;           //現在手先位置・姿勢未入力
  /*  
   *  frm_ctrl code end
   */
   
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RMRC::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMRC::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMRC::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RMRC::onActivated(RTC::UniqueId ec_id)
{
  /*
   *  frm_ctrl code start
   */
  m_com_move.ref_flag = 0;
  m_cur_flag = 0;
  state_old = m_com_frm_ctrl.state;
  /*
   *  frm_ctrl code start
   */

  /*  
   *  mixer code start
   */
  j_list = new double[6*(m_n_size+1)];
  /*  
   *  mixer code end
   */

  /*  
   *  jinv code start
   */
  if((o_n_size != m_n_size) || (o_m_max != m_m_max)) {
    o_n_size = m_n_size;
    o_m_max = m_m_max;
    m_vel.data.length(o_n_size); // angular vel of o_n_size joints
    j_inv = new SVD(o_m_max,o_n_size);
  }
  /*  
   *  jinv code end
   */
   
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RMRC::onDeactivated(RTC::UniqueId ec_id)
{
  /*
   *  frm_ctrl code start
   */
  m_com_move.ref_flag = 0;
  m_cur_flag = 0;
  /*
   *  frm_ctrl code end
   */

  /*  
   *  mixer code start
   */
  delete [] j_list;
  /*  
   *  mixer code end
   */

  return RTC::RTC_OK;
}
/*
RTC::ReturnCode_t RMRC::onExecute(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t RMRC::onAborting(RTC::UniqueId ec_id)
{
  /*
   *  frm_ctrl code start
   */
  m_com_move.ref_flag = 0;
  m_cur_flag = 0;
  /*
   *  frm_ctrl code end
   */

  return RTC::RTC_OK;
}


RTC::ReturnCode_t RMRC::onError(RTC::UniqueId ec_id)
{
  /*
   *  frm_ctrl code start
   */
  m_com_move.ref_flag = 0;
  m_cur_flag = 0;
  /*
   *  frm_ctrl code end
   */

  return RTC::RTC_OK;
}


RTC::ReturnCode_t RMRC::onReset(RTC::UniqueId ec_id)
{
  /*
   *  frm_ctrl code start
   */
  m_com_move.ref_flag = 0;
  m_cur_flag = 0;
  /*
   *  frm_ctrl code end
   */

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RMRC::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMRC::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMRC::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMRC::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMRC::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void RMRCInit(RTC::Manager* manager)
  {
    coil::Properties profile(rmrc_spec);
    manager->registerFactory(profile,
                             RTC::Create<RMRC>,
                             RTC::Delete<RMRC>);
  }
  
};


